-----------------------------------------------------------------------------
-- |
-- Module      :  Decl
-- Project     :  HBC
-- Copyright   :  (c) Hal Daume III
-- License     :  Open
-- 
-- Maintainer  :  Hal Daume III <me@hal3.name>
-- Stability   :  stable
-- Portability :  portable
--
-- The Decl module: implements the base declarations for the HBC
-- system, plus various utilities for traversing and extracting information
-- from these declarations
--
-----------------------------------------------------------------------------

module Decl where

import Data.Maybe

type VAR   = String
type DIST  = String
data CONST = INT Integer | REAL Double                 deriving (Eq, Ord)
data DECL  = DECL  LHS RHS [RANGE]                     deriving (Eq, Ord)
data LHS   = LHS   VAR [VAR]                           deriving (Eq, Ord)
data RHS   = RHS   DIST [EXPR] LIFT                    deriving (Eq, Ord)  -- lift cannot be specified by grammar; is a result of Lift.lift
data RANGE = RANGE VAR EXPR EXPR                       deriving (Eq, Ord)
data EXPR  = VAR   VAR
           | IND   VAR [EXPR]
           | FN    VAR [EXPR]    -- fold OP in here
           | SUMPP Bool RANGE EXPR
           | CONST CONST                               deriving (Eq, Ord)
data MOD   = MOD [DECL]                                deriving (Eq, Ord)
type LIFT  = [(EXPR,EXPR)]  -- initially (VAR,EXPR), but reindex may change the first to an expr

-- Make everything an instance of show with some reasonable description:
instance Show CONST where 
  showsPrec i (INT  j) = showsPrec i j
  showsPrec i (REAL j) = showsPrec i j

instance Show DECL where
  showsPrec i (DECL lhs rhs range) = showsPrec i lhs . showString " ~ " . showsPrec i rhs . showsRange range
    where
      showsRange [] = id
      showsRange (x:xs) = showString " , " . showsPrec i x . showsRange xs

instance Show LHS where
  showsPrec i (LHS v []) = showString v
  showsPrec i (LHS v (x:l)) = showString v . showString "_{" . showString x . showsIx l . showChar '}'
    where
      showsIx [] = id
      showsIx (x:xs) = showChar ',' . showString x . showsIx xs

instance Show RHS where
  showsPrec i (RHS d [] []) = showString d . showString "()"
  showsPrec i (RHS d (x:l) lift) = liftA lift . showString d . showChar '(' . showsPrec i x . showsPa l . showChar ')' . liftB lift
    where
      showsPa [] = id
      showsPa (x:xs) = showString ", " . showsPrec i x . showsPa xs

      liftA [] = id
      liftA ((new,old):xs) = showsPrec i (SUMPP True (RANGE (show new) (VAR "?") (VAR "?")) (VAR "")) . liftA xs

      liftB [] = id
      liftB ((new,old):xs) = showString "^{" . showsPrec i new . showChar '=' . showsPrec i old . showChar '}'

instance Show RANGE where
  showsPrec i (RANGE v lo hi) = showString v . showString " \\in [" . showsPrec i lo . showChar ',' . showsPrec i hi . showChar ']'

instance Show EXPR where
  showsPrec i (VAR v) = showString v
  showsPrec i (IND v []) = showString v
  showsPrec i (IND v (x:l)) = showString v . showString "_{" . showsPrec i x . showsIx l . showChar '}'
    where
      showsIx [] = id
      showsIx (x:xs) = showChar ',' . showsPrec i x . showsIx xs
  showsPrec i (FN v (x:l)) = showString v . showChar '(' . showsPrec i x . showsIx l . showChar ')'
    where
      showsIx [] = id
      showsIx (x:xs) = showString ", " . showsPrec i x . showsIx xs
  showsPrec i (SUMPP prod (RANGE v lo hi) e) = showChar '\\' . (if prod then showString "prod" else showString "sum") . showString "_{" . showString v . showChar '=' . showsPrec i lo . showString "}^{" . showsPrec i hi . showString "} " . showsPrec i  e
  showsPrec i (CONST c) = showsPrec i c

instance Show MOD where
  showsPrec i (MOD []) = id
  showsPrec i (MOD (x:l)) = showsPrec i x . showChar '\n' . showsPrec i (MOD l)

-- -----------------------------------------------------------------------------

-- | extract all ids from various types
declIds  (DECL  lhs rhs rl) = lhsIds lhs ++ rhsIds rhs ++ concatMap rangeIds rl
lhsIds   (LHS   v vl)       = map VAR (v:vl)
rhsIds   (RHS   _ pa lift)  = concatMap exprIds pa ++ concatMap liftIds lift
rangeIds (RANGE v lo hi)    = (VAR v) : exprIds lo ++ exprIds hi
exprIds  (VAR   v)          = [VAR v]
exprIds  (IND   v sub)      = (IND v sub) : concatMap exprIds sub
exprIds  (FN    f pa)       = concatMap exprIds pa
exprIds  (SUMPP _ r e)      = rangeIds r ++ exprIds e
exprIds  (CONST _)          = []
modIds   (MOD   decls)      = concatMap declIds decls
liftIds  (v,e)              = exprIds v ++ exprIds e

isVar (VAR _) = True
isVar _ = False

exprVars = mapMaybe (\x -> case x of { VAR v -> Just v ; _ -> Nothing }) . exprIds

-- -----------------------------------------------------------------------------

-- | evaluate a function as far as possible (for generating simple expressions)

evalFn :: Maybe CONST -> EXPR -> (Maybe VAR, Maybe CONST)
evalFn c0 (VAR v) = (Just v, c0)
evalFn c0 f@(FN op [a,b]) =
  case (evalFn c0 a, evalFn Nothing b) of
    ((Nothing,cA),(v,cB)) -> (v, evalOp op cA cB)
    ((v,cA),(Nothing,cB)) -> (v, evalOp op cA cB)
    ((Just _,_),(Just _,_)) -> error ("Decl.evalFn: both LHS and RHS have variables in: " ++ show f)
evalFn c0 (IND v []) = (Just v,c0)
evalFn c0 (CONST c) = (Nothing, Just c)
evalFn _ z = error ("Decl.evalFn: cannot evaluate: " ++ show z)

evalOp "+" a Nothing = a
evalOp "+" Nothing a = a
evalOp "-" a Nothing = a
evalOp "-" Nothing (Just (INT  i)) = Just (INT  (-i))
evalOp "-" Nothing (Just (REAL r)) = Just (REAL (-r))
evalOp "+" (Just (INT  i)) (Just (INT  j)) = Just (INT  (i+j))
evalOp "+" (Just (INT  i)) (Just (REAL j)) = Just (REAL ((fromIntegral i)+j))
evalOp "+" (Just (REAL i)) (Just (INT  j)) = Just (REAL (i+(fromIntegral j)))
evalOp "+" (Just (REAL i)) (Just (REAL j)) = Just (REAL (i+j))
evalOp "-" (Just (INT  i)) (Just (INT  j)) = Just (INT  (i-j))
evalOp "-" (Just (INT  i)) (Just (REAL j)) = Just (REAL ((fromIntegral i)-j))
evalOp "-" (Just (REAL i)) (Just (INT  j)) = Just (REAL (i-(fromIntegral j)))
evalOp "-" (Just (REAL i)) (Just (REAL j)) = Just (REAL (i-j))
evalOp op  a b = error ("Decl.evalOp: cannot evalute op '" ++ op ++ "' applied to " ++ show (a,b))

